#!/bin/sh

export BASE_DIR="$(cd $(dirname $0); pwd)"
top_dir="$BASE_DIR/.."

if test "$NO_MAKE" != "yes"; then
    make -C ${top_dir} > /dev/null || exit 1
fi

. "${top_dir}/config.sh"

test_suite_names="mroonga_storage,mroonga_wrapper"
source_mysql_test_dir="${MYSQL_SOURCE_DIR}/mysql-test"
build_mysql_test_dir="${MYSQL_BUILD_DIR}/mysql-test"
source_test_suites_dir="${source_mysql_test_dir}/suite"
build_test_suites_dir="${build_mysql_test_dir}/suite"
build_test_include_dir="${build_mysql_test_dir}/include"
case "${MYSQL_VERSION}" in
    5.1.*)
	plugins_dir="${MYSQL_BUILD_DIR}/lib/mysql/plugin"
	if ! test -d "${build_test_suites_dir}"; then
	    mkdir -p "${build_test_suites_dir}"
	fi
	;;
    *)
	if ! test -d "${build_test_suites_dir}"; then
	    ln -s "${source_test_suites_dir}" "${build_test_suites_dir}"
	fi
	maria_storage_dir="${MYSQL_SOURCE_DIR}/storage/maria"
	if test -d "${maria_storage_dir}"; then
	    mariadb="yes"
	else
	    mariadb="no"
	fi
	if test "${mariadb}" = "yes"; then
	    ln -s "${top_dir}" "${MYSQL_BUILD_DIR}/plugin/mroonga"
	    plugins_dir=
	else
	    plugins_dir="${MYSQL_SOURCE_DIR}/lib/plugin"
	fi
	;;
esac

same_link_p()
{
    src=$1
    dest=$2
    if test -L "$dest" -a "$(readlink "$dest")" = "$src"; then
	return 0
    else
	return 1
    fi
}

local_mroonga_mysql_test_include_dir="${BASE_DIR}/sql/include"
for test_include_name in $(ls $local_mroonga_mysql_test_include_dir | grep '\.inc$'); do
    local_mroonga_mysql_test_include="${local_mroonga_mysql_test_include_dir}/${test_include_name}"
    mroonga_mysql_test_include="${build_test_include_dir}/${test_include_name}"
    if ! same_link_p "${local_mroonga_mysql_test_include}" \
			"${mroonga_mysql_test_include}"; then
	rm -f "${mroonga_mysql_test_include}"
        ln -s "${local_mroonga_mysql_test_include}" \
	    "${mroonga_mysql_test_include}"
    fi
done

for test_suite_name in $(echo $test_suite_names | sed -e 's/,/ /g'); do
    local_mroonga_mysql_test_suite_dir="${BASE_DIR}/sql/suite/${test_suite_name}"
    mroonga_mysql_test_suite_dir="${build_test_suites_dir}/${test_suite_name}"
    if ! same_link_p "${local_mroonga_mysql_test_suite_dir}" \
			"${mroonga_mysql_test_suite_dir}"; then
	rm -f "${mroonga_mysql_test_suite_dir}"
	ln -s "${local_mroonga_mysql_test_suite_dir}" \
	    "${mroonga_mysql_test_suite_dir}"
    fi
done

innodb_test_suite_dir="${build_test_suites_dir}/innodb"
mroonga_wrapper_innodb_test_suite_dir="${build_test_suites_dir}/mroonga_wrapper_innodb"
mroonga_wrapper_innodb_include_dir="${mroonga_wrapper_innodb_test_suite_dir}/include/"
if test "$0" -nt "$(dirname "${mroonga_wrapper_innodb_test_suite_dir}")"; then
    rm -rf "${mroonga_wrapper_innodb_test_suite_dir}"
fi
if ! test -d "${mroonga_wrapper_innodb_test_suite_dir}"; then
    cp -rp "${innodb_test_suite_dir}" "${mroonga_wrapper_innodb_test_suite_dir}"
    mkdir -p "${mroonga_wrapper_innodb_include_dir}"
    cp -rp "${build_test_include_dir}"/innodb[-_]*.inc \
	"${mroonga_wrapper_innodb_include_dir}"
    ruby -i'' \
	-pe "\$_.gsub!(/\\bengine\\s*=\\s*innodb\\b([^;\\n]*)/i,
                       \"ENGINE=mroonga\\\1 COMMENT='ENGINE \\\"InnoDB\\\"'\")
             \$_.gsub!(/\\b(storage_engine\\s*=\\s*)innodb\\b([^;\\n]*)/i,
                       \"\\\1mroonga\")
             \$_.gsub!(/^(--\\s*source\\s+)(include\\/innodb)/i,
                       \"\\\1suite/mroonga_wrapper_innodb/\\\2\")
            " \
	${mroonga_wrapper_innodb_test_suite_dir}/r/*.result \
	${mroonga_wrapper_innodb_test_suite_dir}/t/*.test \
	${mroonga_wrapper_innodb_test_suite_dir}/include/*.inc
    sed -i'' \
	-e '1 i --source include/have_mroonga.inc' \
	${mroonga_wrapper_innodb_test_suite_dir}/t/*.test
fi

if test -n "${plugins_dir}"; then
    if test -d "${top_dir}/.libs"; then
	make -C ${top_dir} \
	    install-pluginLTLIBRARIES \
	    plugindir=${plugins_dir} > /dev/null || \
	    exit 1
    else
	cp "${top_dir}/ha_mroonga.so" "${plugins_dir}" || exit 1
    fi
fi

(cd "$build_mysql_test_dir" && \
    ./mysql-test-run.pl \
    --no-check-testcases \
    --parallel=4 \
    --retry=1 \
    --suite="${test_suite_names}" \
    --force \
    "$@")
